/* ********************************************************************* *
   mod_auth_tkt
   MD5 based ticket-type user authentication with optional timeout and 
   authentication of user supplied data 
 * ********************************************************************* */

/* =======================================================================
 * Portions Copyright (c) 2001-2002 Open Fusion Pty. Ltd. Australia.
 * All rights reserved. 
 * Portions Copyright (c) 2000 Liquid Digital Information Systems, Inc. 
 * All rights reserved. 
 *
 * This software was written for Open Fusion Pty. Ltd. (Australia) by 
 * Gavin Carr, and for Liquid Digital Information Systems, Inc. 
 * by Raimondas Kiveris. It is hereby contributed to the Apache Software 
 * Foundation for distribution under the Apache license, as follows.
 *	
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 * 
 * 3. All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by the Apache Group
 *  for use in the Apache HTTP server project (http://www.apache.org/)."
 * 
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *  endorse or promote products derived from this software without
 *  prior written permission. For written permission, please contact
 *  apache@apache.org.
 * 
 * 5. Products derived from this software may not be called "Apache"
 *  nor may "Apache" appear in their names without prior written
 *  permission of the Apache Group.
 * 
 * 6. Redistributions of any form whatsoever must retain the following
 *  acknowledgment:
 *  "This product includes software developed by the Apache Group
 *  for use in the Apache HTTP server project (http://www.apache.org/)."
 * 
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * =======================================================================
 *	
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 * 
 * ======================================================================
 */


#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#include "tkt_digest.h"

#include "test_char.h"
#define TEST_CHAR(c, f) (test_char_table[(unsigned)(c)] & (f))
static const char c2x_table[] = "0123456789abcdef";

#include "mod_auth_tkt.h"

module MODULE_VAR_EXPORT tkt_auth_module;


/*****************************************************************/



/* initializer */

void tkt_auth_init (server_rec *s, pool *p)
{
    ap_add_version_component("TKTAuth/" TKT_AUTH_VERSION);
}

/* -------------------------------------- */


/* dir config creator */

static void * 
create_tkt_auth_config(pool *p, char* path)
{
  tkt_auth_dir_conf *new =
    (tkt_auth_dir_conf *) ap_palloc(p, sizeof(tkt_auth_dir_conf));

  new->directory = path;
  new->login_url = NULL;
  new->timeout_url = NULL;
  new->unauth_url = NULL;
  new->auth_domain = NULL;
  new->auth_token = NULL;
  new->auth_cookie_name = NULL;
  new->back_cookie_name = NULL;
  new->ignore_ip = 0;
  new->timeout_sec = 0;
  new->timeout_refresh = .5;

  return (void *) new;  
}

/* -------------------------------------- */



/* dir merger */

static void * 
merge_tkt_auth_config(pool *p, void* parent_dirv, void* subdirv)
{
  tkt_auth_dir_conf *parent = (tkt_auth_dir_conf *) parent_dirv;
  tkt_auth_dir_conf *subdir = (tkt_auth_dir_conf *) subdirv;
  tkt_auth_dir_conf *new =
    (tkt_auth_dir_conf *) ap_palloc(p, sizeof(tkt_auth_dir_conf));

  new->directory = (subdir->directory) ? subdir->directory : parent->directory;
  new->login_url = (subdir->login_url) ? subdir->login_url : parent->login_url;
  new->timeout_url = (subdir->timeout_url) ? subdir->timeout_url : parent->timeout_url;
  new->unauth_url = (subdir->unauth_url) ? subdir->unauth_url : parent->unauth_url;
  new->auth_domain = (subdir->auth_domain) ? subdir->auth_domain : parent->auth_domain;
  new->auth_token = (subdir->auth_token) ? subdir->auth_token : parent->auth_token;
  new->auth_cookie_name = (subdir->auth_cookie_name) ? subdir->auth_cookie_name : parent->auth_cookie_name;
  new->back_cookie_name = (subdir->back_cookie_name) ? subdir->back_cookie_name : parent->back_cookie_name;
  new->ignore_ip = (subdir->ignore_ip) ? subdir->ignore_ip : parent->ignore_ip;
  new->timeout_sec = (subdir->timeout_sec) ? subdir->timeout_sec : parent->timeout_sec;
  new->timeout_refresh = (subdir->timeout_refresh) ? subdir->timeout_refresh : parent->timeout_refresh;

  return (void *) new;
}

/* -------------------------------------- */



/* server config */

static void *
create_tkt_auth_serv_config(pool *p, server_rec* s)
{
  tkt_auth_serv_conf *new =
    (tkt_auth_serv_conf *) ap_palloc(p, sizeof(tkt_auth_serv_conf));
  new->md5_key = NULL;

  return (void *) new;
} 

/* -------------------------------------- */



/* merge server config */

/* -------------------------------------- */



/* command table */

static const command_rec tkt_auth_cmds[] =
{
  {"TKTAuthLoginURL", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, login_url),
   OR_AUTHCFG, TAKE1, "where to redirect to if authentication fails."},
  {"TKTAuthTimeoutURL", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, timeout_url),
   OR_AUTHCFG, TAKE1, "where to redirect to if cookie times-out."},
  {"TKTAuthUnauthURL", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, unauth_url),
   OR_AUTHCFG, TAKE1, "where to redirect to if valid user without required token."},
  {"TKTAuthIgnoreIP", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, ignore_ip),
   OR_AUTHCFG, TAKE1, "ip address to use instead of remote addr."},
  {"TKTAuthCookieName", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, auth_cookie_name),
   OR_AUTHCFG, TAKE1, "name to use for ticket cookie."},
  {"TKTAuthBackCookieName", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, back_cookie_name),
   OR_AUTHCFG, TAKE1, "name to use for back cookie (null for none)."},
  {"TKTAuthDomain", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, auth_domain),
   OR_AUTHCFG, TAKE1, "domain to use in cookies."},
  {"TKTAuthToken", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, auth_token),
   OR_AUTHCFG, TAKE1, "token required to access this area (NULL for none)."},
  {"TKTAuthTimeoutMin", set_tkt_auth_timeout, 
   NULL, OR_AUTHCFG, TAKE1, "cookie inactivity timeout in minutes."},
  {"TKTAuthTimeoutRefresh", set_tkt_auth_timeout_refresh, 
   NULL, OR_AUTHCFG, TAKE1, "cookie timeout refresh flag."},
  {"TKTAuthSecret", setup_md5key, NULL, 
   RSRC_CONF, TAKE1, "secret key to use in MD5 digest."},

  {NULL},
};

/* -------------------------------------- */


/* command specific functions */


static const char *
setup_md5key (cmd_parms *cmd, void *dir_cfg, char *param)
{
  tkt_auth_serv_conf *serv_cfg = (tkt_auth_serv_conf *)
    ap_get_module_config(cmd->server->module_config, &tkt_auth_module);

  serv_cfg->md5_key = param;
  return NULL;
}

static const char *
set_tkt_auth_timeout(cmd_parms *cmd, void *cfg, char *param)
{
  tkt_auth_dir_conf *dir_cfg = (tkt_auth_dir_conf *)cfg;
  
  long int minutes = strtol (param, (char **) NULL, 10);
  
  if (minutes < 0) 
    return ("Timeout must be positive");
  
  if (minutes == LONG_MAX) 
    return ("Integer overflow or invalid number");
  
  dir_cfg->timeout_sec = minutes * 60;
  
  return (NULL);
}

static const char *
set_tkt_auth_timeout_refresh(cmd_parms *cmd, void *cfg, char *param)
{
  tkt_auth_dir_conf *dir_cfg = (tkt_auth_dir_conf *)cfg;
  
  double refresh = atof(param);

  if (refresh < 0 || refresh > 1) 
    return ("Refresh flag must be between 0 and 1");
  
  dir_cfg->timeout_refresh = refresh;
  
  return (NULL);
}

/* -------------------------------------- */


/* handlers */
/* -------------------------------------- */



/* filename translation */
/* -------------------------------------- */



/* check_user_id */

static int 
authenticate_tkt_user (request_rec *r)
{
  tkt_auth_dir_conf *tkt_cfg = 
    ap_get_module_config(r->per_dir_config, &tkt_auth_module);
  tkt_auth_serv_conf *tkt_scfg =
    ap_get_module_config(r->server->module_config, &tkt_auth_module);

  char *cookie; //, *cookie_path;
  tkt_cookie * parsed;
  int rc;
  unsigned int now = time(NULL);
  char *digest;
  char *remote_ip;
  
  /* Default cookie name */
  if (! tkt_cfg->auth_cookie_name) {
      tkt_cfg->auth_cookie_name = AUTH_COOKIE_NAME;
  }

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT: tkt_cfg: login_url %s, auth_cookie_name %s, auth_domain %s, auth_token %s, timeout_sec %d", 
    tkt_cfg->login_url, 
    tkt_cfg->auth_cookie_name,
    tkt_cfg->auth_domain,
    tkt_cfg->auth_token,
    tkt_cfg->timeout_sec);

  if (! tkt_cfg->login_url) {	/* module is not configured here */
    return DECLINED;
  }

  
  if (! tkt_scfg->md5_key) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "TKT: TKTAuthSecret missing");
    return SERVER_ERROR;
  }

  //  cookie_path = tkt_cfg->directory+strlen(ap_document_root(r));

  remote_ip = (tkt_cfg->ignore_ip != NULL && strncmp(tkt_cfg->ignore_ip,"0",1) != 0) ? "0.0.0.0" : r->connection->remote_ip;

  cookie = find_our_cookie(r, tkt_cfg);
  
  if (cookie == NULL) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT: no cookie, redirecting to %s", 
		  tkt_cfg->login_url);
    
    return (redirect(r, tkt_cfg->login_url, tkt_cfg->back_cookie_name));
  }
  
  parsed = ap_palloc(r->pool, sizeof(tkt_cookie));
  
  rc = parse_cookie(r, &cookie, parsed);
  
  if (rc != 0) {
    ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "TKT: corrupted cookie <%s>", 
		  cookie);
    
    return (redirect(r, tkt_cfg->login_url, tkt_cfg->back_cookie_name));
  }

  /*
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
      "TKT: parsed cookie: uid %s, timestamp %d, tokens %s, data %s",
      parsed->uid, parsed->timestamp, parsed->tokens, parsed->user_data);
  */

  // Check timeout
  if (tkt_cfg->timeout_sec > 0) {
    
    if (now > tkt_cfg->timeout_sec + parsed->timestamp) {
      ap_log_rerror(APLOG_MARK, APLOG_INFO, r, 
		    "TKT: cookie timed out: now %d, timeout: %ld, tstamp: %d", 
		    now, tkt_cfg->timeout_sec, parsed->timestamp);
      
      return redirect(r, tkt_cfg->timeout_url ? 
                      tkt_cfg->timeout_url : tkt_cfg->login_url,
                      tkt_cfg->back_cookie_name);
    }
  }

  digest = tkt_digest(remote_ip,        /* digest is malloc'ed in tkt_digest */
		      tkt_scfg->md5_key, 
		      parsed->uid, 
                      parsed->tokens,
		      parsed->user_data, 
		      parsed->timestamp);

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
      "TKT: digest check on cookie %-32.32s, digest %s, "
      "ip %s, key %s, uid %s, timestamp %d, tokens %s, data %s",
      cookie, digest,
      remote_ip, tkt_scfg->md5_key, parsed->uid, 
      parsed->timestamp, parsed->tokens, parsed->user_data);

  if (!memcmp(cookie, digest, MD5_DIGEST_SZ)) { /* user is authenticated */
    char buf[640];
    char *cookie;
    char *domain = tkt_cfg->auth_domain;
    if (!domain) {
      domain = strchr(ap_get_server_name(r), '.');
    }
    
    free (digest);
    
    /* check for auth_token, if required */
    if (tkt_cfg->auth_token && strcmp(tkt_cfg->auth_token,"NULL") != 0) {
      int unauth = 1;
      char buf[640];
      char *b = buf;
      char *next;
      int token_len = strlen(tkt_cfg->auth_token);
      snprintf(buf, 640, "%s", parsed->tokens);

      /* Check tokens */
      while (next = strsep(&b, ",")) {
          if (!strncmp(tkt_cfg->auth_token, next, token_len)) {
            unauth = 0;
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
              "TKT: checking auth_token '%s' against tokens '%s' - found",
              tkt_cfg->auth_token, parsed->tokens);
            break;
        }
      }

      /* Redirect if not authorised */
      if (unauth) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
          "TKT: checking auth_token '%s' against tokens '%s' - not found, redirecting",
          tkt_cfg->auth_token, parsed->tokens);
        return (redirect(r, tkt_cfg->unauth_url ? 
                            tkt_cfg->unauth_url : tkt_cfg->login_url, 
                            tkt_cfg->back_cookie_name));
      }
    }

    /* Refresh the cookie if using inactivity timeouts */
    if (tkt_cfg->timeout_sec > 0 && tkt_cfg->timeout_refresh > 0) {

      /* The timeout refresh is a double between 0 and 1, signifying what
       * proportion of the timeout should be left before we refresh i.e. 
       * 0 means never refresh (hard timeouts); 1 means always refresh;
       * .33 means only refresh after two-thirds of the timeout period has 
       * elapsed. */ 
      int elapsed = now - parsed->timestamp;
      if (elapsed > (1 - tkt_cfg->timeout_refresh) * tkt_cfg->timeout_sec) {
      
        /* Reconstruct the cookie  */
        digest = tkt_digest(remote_ip,
			    tkt_scfg->md5_key, 
			    parsed->uid, 
			    parsed->tokens, 
			    parsed->user_data, 
			    now);
      
        memcpy (buf, digest, 32);
        if (parsed->tokens) {
            sprintf(&buf[32], "%08x%s%c%s%c%s", 
                    now, 
                    parsed->uid, 
                    SEPARATOR,
                    parsed->tokens,
                    SEPARATOR,
                    parsed->user_data);
        }
        else {
            sprintf(&buf[32], "%08x%s%c%s", 
                    now, 
                    parsed->uid, 
                    SEPARATOR,
                    parsed->user_data);
        }
        
        {
	  char *copy = strdup (buf);
	  ap_base64encode(buf, copy, strlen(copy));
	  free(copy);
        }
      
        cookie =	
//	  ((ap_get_server_port(r) == 80) ? 
	   ap_psprintf(r->pool, 
                       "%s=%s; path=/; domain=%s", 
		       tkt_cfg->auth_cookie_name,
		       buf,
		       //		     cookie_path,
		       domain);
/*
	   :
	   ap_psprintf(r->pool, 
                       "%s=%s:%d; path=/; domain=%s", 
		       tkt_cfg->auth_cookie_name,
		       buf,
		       ap_get_server_port(r),
		       //		     cookie_path,
		       domain));
*/
      
        ap_table_setn (r->err_headers_out, "Set-Cookie", cookie);
        free(digest);
      }
    }
    
    r->connection->user = parsed->uid;
    r->connection->ap_auth_type = "Basic";	/* for other modules */
    
    ap_table_set(r->subprocess_env, REMOTE_USER_DATA_ENV, parsed->user_data);

    add_query_string(r, parsed->user_data);
   
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
        "TKT: user '%s' authenticated, tokens '%s', data '%s'", 
            parsed->uid, parsed->tokens, parsed->user_data);
 
    return (OK);
  }

  ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, 
        "TKT: someone tampered with cookie <%s>", 
		cookie);
  
  free(digest);
  return (redirect(r, tkt_cfg->login_url, tkt_cfg->back_cookie_name));
}

/* -------------------------------------- */


/* check auth */
/* -------------------------------------- */


/* check access */
/* -------------------------------------- */


/* type_checker */
/* -------------------------------------- */


/* fixups */
/* -------------------------------------- */


/* logger */
/* -------------------------------------- */


/* header parser */
/* -------------------------------------- */


/* chitkt_init */
/* -------------------------------------- */


/* chitkt_exit */
/* -------------------------------------- */


/* post read-request */
/* -------------------------------------- */







/* misc functions */

static int 
parse_cookie(request_rec *r, char **magic, tkt_cookie *parsed)
{
  char *cookie = *magic;
  int len = strlen (cookie);
  int sepidx, sep2idx;
  
  if (len <= MD5andTSTAMP) {
    return (1); /* not our cookie */
  }
  
  /* see if there is a uid/data separator */
  sepidx = ap_ind(cookie, SEPARATOR);
  
  if (sepidx == -1) {	/* cookie either in Base64 format, 
			 * has URI hex escapes or is corrupted */
    if (strstr(cookie, SEPARATOR_HEX) ) { /* hex('!')  */
      ap_unescape_url(cookie);
      
      sepidx = ap_ind(cookie, SEPARATOR);
    }
    else {
      char *buf = (char *) ap_palloc(r->pool, len);  /* Base64 encoded string is 
						      * longer than original */
      ap_base64decode(buf, cookie);
      
      sepidx = ap_ind (buf, SEPARATOR);
      
      if (sepidx == -1) {
	return (1);				  /* not our cookie */
      }
      
      cookie = *magic = buf;
    }
    
    len = strlen(cookie);
  }
  
  /* get the user id */
  parsed->uid = ap_palloc(r->pool, sepidx - MD5andTSTAMP + 1);
  memcpy(parsed->uid, &cookie[MD5andTSTAMP], sepidx - MD5andTSTAMP);
  parsed->uid[sepidx - MD5andTSTAMP] = '\0';
  
  /* check for tokens */
  sep2idx = ap_ind (&cookie[sepidx+1], SEPARATOR);
  if (sep2idx == -1) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT parse_cookie: no tokens");
      parsed->tokens = ap_palloc(r->pool, 1);
      *parsed->tokens = '\0';
  }
  else {
      // swap sepidx and sep2idx
      int tmp = sepidx;
      sepidx = tmp + sep2idx + 1;
      sep2idx = tmp;
//    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT parse_cookie: tokens found - sep2=%d, sep=%d, len=%d", sep2idx, sepidx, len);
      parsed->tokens = ap_palloc(r->pool, sepidx - sep2idx);
      snprintf (parsed->tokens, sepidx-sep2idx, "%s", &cookie[sep2idx+1]);
//    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT parse_cookie tokens: %s", parsed->tokens);
  }

  /* user data sent in a cookie */
  parsed->user_data = ap_palloc(r->pool, len - sepidx);
  snprintf (parsed->user_data, len-sepidx, "%s", &cookie[sepidx+1]);
  
  /* and timestamp */
  sscanf(&cookie[MD5_DIGEST_SZ], "%8x", &(parsed->timestamp));
  
  return (0);
}


static int 
cookie_match(void *result, const char *key, const char *cook)
{
  char *value;
  cookie_res * cr = (cookie_res *) result;
  
  if (cook != NULL) {
    // ap_log_rerror(APLOG_MARK, APLOG_DEBUG, cr->r, "TKT cookie_match, key %s against <%s> (name=%s)", key, cook, cr->cookie_name);
    
    if (value = strstr(cook, cr->cookie_name)) {
      char *cookiebuf, *cookieend;
      
      value += strlen(cr->cookie_name) + 1;  /* Skip over the '=' */
      cookiebuf = ap_pstrdup(cr->r->pool, value);
      cookieend = strchr(cookiebuf, ';');
      if (cookieend)
	*cookieend = '\0';      /* Ignore anything after a ; */
      
      cr->cookie = cookiebuf;
//    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, cr->r, "TKT cookie_match: found");
      return(0);
    }
  }
//ap_log_rerror(APLOG_MARK, APLOG_DEBUG, cr->r, "TKT cookie_match: NOT found");
  return (1);
}


static char * 
find_our_cookie(request_rec *r, tkt_auth_dir_conf *tkt_cfg)
{
  cookie_res *cr = ap_palloc(r->pool, sizeof(cookie_res));
  
  cr->r = r;
  cr->cookie = NULL;
  cr->cookie_name = tkt_cfg->auth_cookie_name;
  
  ap_table_do(cookie_match, (void *) cr, r->headers_in, "Cookie", NULL);
  
  return (cr->cookie);
}


static ap_inline unsigned char *c2x(unsigned what, unsigned char *where)
{
  *where++ = '%';
  *where++ = c2x_table[what >> 4];
  *where++ = c2x_table[what & 0xf];
  return where;
}

static char * 
escape_str(pool *p, const char *segment)
{
  char *copy = ap_palloc(p, 3 * strlen(segment) + 1);
  const unsigned char *s = (const unsigned char *)segment;
  unsigned char *d = (unsigned char *)copy;
  unsigned c;
  
  while ((c = *s)) {
    if (TEST_CHAR(c, T_ESCAPE_PATH_SEGMENT) || c == '=' || c == '&') {
      d = c2x(c, d);
    }
    else {
      *d++ = c;
    }
    ++s;
  }
  *d = '\0';
  return copy;
}


static int 
redirect(request_rec *r, char *location, char *back_cookie_name)
{
  /* set cookie with currently accessed location */
  char *domain = strchr(ap_get_server_name(r), '.');
  char *url;
  char *cookie; //, *cookie_path;
  char *back;
  char sep;
  int sz;

  //  tkt_auth_dir_conf *tkt_cfg = 
  //    ap_get_module_config(r->per_dir_config, &tkt_auth_module);
  
  //  cookie_path = tkt_cfg->directory+strlen(ap_document_root(r));

  /* Set back URL */
  back = ((ap_get_server_port(r) == 80) ? 
       ap_psprintf(r->pool, 
                   "%s%s", 
                   ap_get_server_name(r), 
                   escape_str(r->pool, r->unparsed_uri))
       :
       ap_psprintf(r->pool, 
                   "%s:%d%s", 
                   ap_get_server_name(r), 
                   ap_get_server_port(r), 
                   escape_str(r->pool, r->unparsed_uri))); 
  
  /* Set back cookie if name is not null */
  if (back_cookie_name) {
    cookie = ap_psprintf(r->pool, 
                   "%s=%s; path=/; domain=%s", 
                   back_cookie_name,
                   back,
                   domain);
    ap_table_setn(r->err_headers_out, "Set-Cookie", cookie);
    
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
          "TKT back cookie: |%s:%s|", back_cookie_name, cookie);
  
    sz = 1 + strlen(location) + 1;
    url = ap_palloc(r->pool, sz);
    snprintf (url, sz, "%s", location);
  }

  /* If back_cookie_name not set, add a back parameter to location */
  else {
    sz = 1 + strlen(location) + 6 + strlen(back) + 1;
    url = ap_palloc(r->pool, sz);
    sep = (strchr(location, '?')) ? '&' : '?';
    snprintf (url, sz, "%s%cback=%s", location, sep, back);
  }

/*
  if (location[0] == '/') { 
    snprintf (url, sz, "%s", location);
  }
  else {
    snprintf (url, sz, "/%s", location);
  }
*/
  
  ap_table_setn(r->headers_out, "Location", url);
  
  return REDIRECT;
}


static void
add_query_string(request_rec *r, char *data)
{
  if (r->args && r->args[0] != '\0') {
    char *start;
    char *end;
		
    /* here we have to prevent unauthorized REMOTE_USER_DATA
     * from being passed as args */
    if ((start = strstr(r->args, REMOTE_USER_DATA_TAG)) != NULL) {
      if (start == r->args) {
	end = strstr(start, "&");
	if (end == NULL) {
	  r->args = NULL;
	}
	else {
	  r->args = ++end;
	}
      } /* end of if (start == r->args) */
      else {
	end = strstr(start, "&");
	*(--start) = '\0';
	if (end != NULL) {
	  r->args = ap_pstrcat(r->pool, r->args, end, NULL);
	}
      }
    }
  }
  
  return ;

  if (r->args && r->args[0] != '\0') {
    r->args = ap_pstrcat(r->pool, 
			 REMOTE_USER_DATA_TAG "=", 
			 ap_escape_path_segment(r->pool, data), 
			 "&", 
			 r->args, NULL);
  }
  else {
    r->args = ap_pstrcat(r->pool, 
			 REMOTE_USER_DATA_TAG "=", 
			 ap_escape_path_segment(r->pool, data), 
			 NULL);
  }

}


/* -------------------------------------- */



/*****************************************************************/





module MODULE_VAR_EXPORT tkt_auth_module =
{
	STANDARD_MODULE_STUFF,
	tkt_auth_init,		/* initializer */
	create_tkt_auth_config,	/* dir config creator */
	merge_tkt_auth_config,	/* dir merger --- default is to override */
	create_tkt_auth_serv_config,			/* server config */
	NULL,			/* merge server config */
	tkt_auth_cmds,	        /* command table */
	NULL,			/* handlers */
	NULL,			/* filename translation */
	authenticate_tkt_user,	/* check user_id */
	NULL,		        /* check auth */
	NULL,			/* check access */
	NULL,			/* type_checker */
	NULL,			/* fixups */
	NULL,			/* logger */
	NULL,			/* header parser */
	NULL,			/* chitkt_init */
	NULL,			/* chitkt_exit */
	NULL			/* post read-request */
};





